CON

{{
Note:   Current valid received data for all channels is held in the
        curr_frame[] array.  Pulse times that are written to servos
        are stored in servos[] array.  See the constant enumeration
        below for named array positions that correspond to the transmiter.

        For example, if you want to know the analog value of the right gimbal's
        horizontal axis, simply reference the data in "curr_frame[R_GIM_HORZ]".

        All channel data is in the range of 0 -> 4095, corresponding from 0% -> 100%,
        respectively.


  R_GIM_VERT    'curr_frame[0]
  R_GIM_HORZ    'curr_frame[1]         
  L_GIM_HORZ    'curr_frame[2]
  L_GIM_VERT    'curr_frame[3]
  ANLG2         'curr_frame[4]
  ANLG1         'curr_frame[5]
  ANLG4         'curr_frame[6]
  ANLG3         'curr_frame[7]
                '
  R_OVAL        'curr_frame[8]
  L_OVAL        'curr_frame[9]
  TOGGLE1       'curr_frame[10]
  TOGGLE2       'curr_frame[11]
  TOGGLE3       'curr_frame[12]
  TOGGLE4       'curr_frame[13]
  TOGGLE5       'curr_frame[14]
  TOGGLE6       'curr_frame[15]
                '
  R_UP          'curr_frame[16]
  R_RIGHT       'curr_frame[17]
  R_LEFT        'curr_frame[18]
  R_DOWN        'curr_frame[19]
  L_RIGHT       'curr_frame[20]
  L_DOWN        'curr_frame[21]
  L_UP          'curr_frame[22]
  L_LEFT        'curr_frame[23]





}}


  _xinfreq = 5_000_000
  _clkmode = xtal1 + pll16x


  XBEE_DOUT_PIN = 10
  XBEE_DIN_PIN  = 11
  XBEE_BAUD     = 115200
  XBEE_MODE     = %0000

  SERVO_PERIOD  = 20            'in milliseconds

  PACKET_SIZE   = 3 * 8         'in bytes, 3 ADCs x 8 channels per ADC

  NUM_CHANNELS  = 8

  'Propeller pin mappings, these map channels to pin outputs.
  CHANNEL_0     = 14
  CHANNEL_1     = 15
  CHANNEL_2     = 16
  CHANNEL_3     = 17
  CHANNEL_4     = 18
  CHANNEL_5     = 19
  CHANNEL_6     = 20
  CHANNEL_7     = 21

  '=== these channels aren't put to any pins in this implementation  
  CHANNEL_8     = -1
  CHANNEL_9     = -1
  CHANNEL_10    = -1
  CHANNEL_11    = -1
  CHANNEL_12    = -1
  CHANNEL_13    = -1
  CHANNEL_14    = -1
  CHANNEL_15    = -1
  CHANNEL_16    = -1
  CHANNEL_17    = -1
  CHANNEL_18    = -1
  CHANNEL_19    = -1
  CHANNEL_20    = -1
  CHANNEL_21    = -1
  CHANNEL_22    = -1
  CHANNEL_23    = -1

  SERVO_MIN_TIME= 1000          'in microseconds
  SERVO_MAX_TIME= 2000          'in microseconds

  CHANNEL_MIN   = 0
  CHANNEL_MAX   = $0FFF


  REVERSE_CH0   = true
  REVERSE_CH1   = true
  REVERSE_CH2   = false
  REVERSE_CH3   = false
  REVERSE_CH4   = false
  REVERSE_CH5   = false
  REVERSE_CH6   = false
  REVERSE_CH7   = false
  

  'Channel Enumerations - should match how channels are defined in the TX code
  #0
  R_GIM_VERT    'curr_frame[0]
  R_GIM_HORZ    'curr_frame[1]         
  L_GIM_HORZ    'curr_frame[2]
  L_GIM_VERT    'curr_frame[3]
  ANLG2         'curr_frame[4]
  ANLG1         'curr_frame[5]
  ANLG4         'curr_frame[6]
  ANLG3         'curr_frame[7]
                '
  R_OVAL        'curr_frame[8]
  L_OVAL        'curr_frame[9]
  TOGGLE1       'curr_frame[10]
  TOGGLE2       'curr_frame[11]
  TOGGLE3       'curr_frame[12]
  TOGGLE4       'curr_frame[13]
  TOGGLE5       'curr_frame[14]
  TOGGLE6       'curr_frame[15]
                '
  R_UP          'curr_frame[16]
  R_RIGHT       'curr_frame[17]
  R_LEFT        'curr_frame[18]
  R_DOWN        'curr_frame[19]
  L_RIGHT       'curr_frame[20]
  L_DOWN        'curr_frame[21]
  L_UP          'curr_frame[22]
  L_LEFT        'curr_frame[23]
  
    
OBJ

  d     :       "Parallax Serial Terminal"
  xb    :       "Parallax Serial Terminal"
  f     :       "FloatMath"
  crc   :       "CRC32"

VAR

  byte servo_lock

  word curr_frame[PACKET_SIZE]
  word servos[NUM_CHANNELS]
  word in_buf[PACKET_SIZE]

  long fScale, fOffset
  
  long servo_out_stack[100]
  
PUB Go | t, dt, in, i, rxCRC, calcCRC

  d.Start(115_200)
  xb.StartRxTx(XBEE_DIN_PIN, XBEE_DOUT_PIN, XBEE_MODE, XBEE_BAUD)

  if (servo_lock := locknew) == -1
    reboot                                              'something is really wrong if this ever happens

  wordfill(@servos, 1500, NUM_CHANNELS)
  
  cognew(ServoOut, @servo_out_stack)

'  dira[12]~~
'  dira[13]~~

  'do some pre-calculations to map the received input to servo pulse times
  fScale  := FCalcM(CHANNEL_MIN, SERVO_MIN_TIME, CHANNEL_MAX, SERVO_MAX_TIME)
  fOffset := FCalcB(CHANNEL_MIN, SERVO_MIN_TIME, fScale)


  repeat  'main loop

    repeat until xb.CharIn == $AA
    repeat while (in:=xb.CharIn) == $AA    
    if in <> "!"
      next
    if xb.CharIn <> "Y"
      next
'    !outa[12]
    
    'we should be in valid packet data zone here...
    
    repeat i from 0 to constant(PACKET_SIZE-1)      
      in := xb.CharIn                  'rx the low byte first
      in_buf[i] := (xb.CharIn<<8) | in 'rx the high byte, combine, and save

    repeat i from 0 to 3
      rxCRC.byte[i] := xb.CharIn

    calcCRC := crc.DoCRC32(@in_buf, constant(PACKET_SIZE*2), 0)

    'Make sure the CRC checks out. if so, then copy the data to the valid array, curr_frame[].
    if rxCRC == calcCRC
      repeat while lockset(servo_lock)                    'spin here while the lock isn't available
      wordmove(@curr_frame, @in_buf, PACKET_SIZE)
      lockclr(servo_lock)
'      !outa[13]
       
{
    d.home
    repeat i from 0 to constant(PACKET_SIZE-1)
      d.dec(i)
      d.str(string(": "))
      d.dec(curr_frame[i])
      d.chars(" ", 5)
      d.char(13)

    d.char(13)
    d.hex(rxCRC, 8)
    d.char(13)
    d.hex(calcCRC, 8)
}     
    
  
PUB PulseOut(pin, time_us)

  outa[pin]~~
  PulseOutPauseUs(time_us)
  outa[pin]~
  

PUB ServoOut | dt, t, temp, i, ch_t, ch_dt
''
''  Caution: make sure that this loop never blocks.  It does the work of sticking
''           the pulses to the servos, so we want to make sure pulses are always
''           being sent to the servos.  Special care must be taken here.
''


  dira[CHANNEL_0]~~
  dira[CHANNEL_1]~~
  dira[CHANNEL_2]~~
  dira[CHANNEL_3]~~
  dira[CHANNEL_4]~~
  dira[CHANNEL_5]~~
  dira[CHANNEL_6]~~
  dira[CHANNEL_7]~~


  outa[CHANNEL_0]~
  outa[CHANNEL_1]~
  outa[CHANNEL_2]~
  outa[CHANNEL_3]~
  outa[CHANNEL_4]~
  outa[CHANNEL_5]~
  outa[CHANNEL_6]~
  outa[CHANNEL_7]~


  dt := (clkfreq/1000)*SERVO_PERIOD
  t := cnt + dt
  ch_dt := (clkfreq/1_000_000) * 2250

  repeat

    if lockset(servo_lock) <> -1                        'try to check out the lock, skip if not available

      servos[0] := CalcServoPulseTime(curr_frame[L_GIM_VERT])
      servos[1] := CalcServoPulseTime(curr_frame[R_GIM_HORZ])
      servos[2] := CalcServoPulseTime(curr_frame[R_GIM_VERT])
      servos[3] := CalcServoPulseTime(curr_frame[L_GIM_HORZ])
      
      servos[4] := CalcServoPulseTime(curr_frame[ANLG1])
      'servos[4] := constant(1500+(500>>2))              'manually set this channel to a static value, my gain value for a Parallax Elev-8
      servos[5] := CalcServoPulseTime(curr_frame[ANLG2])
      servos[6] := CalcServoPulseTime(curr_frame[ANLG3])
      servos[7] := CalcServoPulseTime(curr_frame[ANLG4])

      lockclr(servo_lock)                               'all done reading, release the lock

      if REVERSE_CH0
        servos[0] := 3000 - servos[0]
      if REVERSE_CH1
        servos[1] := 3000 - servos[1]
      if REVERSE_CH2
        servos[2] := 3000 - servos[2]
      if REVERSE_CH3
        servos[3] := 3000 - servos[3]
      if REVERSE_CH4
        servos[4] := 3000 - servos[4]
      if REVERSE_CH5
        servos[5] := 3000 - servos[5]
      if REVERSE_CH6
        servos[6] := 3000 - servos[6]
      if REVERSE_CH7
        servos[7] := 3000 - servos[7]

    
    'stick the pulses to the servos here! :D
    ch_t := cnt + ch_dt
    
    PulseOut(CHANNEL_0, servos[0])
    waitcnt(ch_t)
    ch_t += ch_dt
    
    PulseOut(CHANNEL_1, servos[1])
    waitcnt(ch_t)
    ch_t += ch_dt

    PulseOut(CHANNEL_2, servos[2])
    waitcnt(ch_t)
    ch_t += ch_dt

    PulseOut(CHANNEL_3, servos[3])
    waitcnt(ch_t)
    ch_t += ch_dt

    PulseOut(CHANNEL_4, servos[4])
    waitcnt(ch_t)
    ch_t += ch_dt

    PulseOut(CHANNEL_5, servos[5])
    waitcnt(ch_t)
    ch_t += ch_dt

    PulseOut(CHANNEL_6, servos[6])
    waitcnt(ch_t)
    ch_t += ch_dt

    PulseOut(CHANNEL_7, servos[7])
    waitcnt(ch_t)
    ch_t += ch_dt

    'safely wait for the remainder of the servo period
    if (cnt - t) => dt
      waitcnt(t)
    else
      t := cnt

      
    'set up for the next servo period  
    t += dt


PUB CalcServoPulseTime(channelVal)
'
''  remember y = mx + b ?

  channelVal := f.FFloat(channelVal)

  return f.FRound(f.FAdd(f.FMul(channelVal, fScale), fOffset))

  
PRI PulseOutPauseUs(duration)

  duration := SERVO_MIN_TIME #> duration <# SERVO_MAX_TIME

  waitcnt(cnt + (clkfreq/1_000_000)*duration)


PRI FCalcM(x1, y1, x2, y2)
  
  x1 := f.FFloat(x1)
  y1 := f.FFloat(y1)
  x2 := f.FFloat(x2)
  y2 := f.FFloat(y2)        
  
  return f.fdiv(f.FSub(y2, y1),f.FSub(x2, x1))

PRI FCalcB(x, y, f_m)

  x := f.FFloat(x)
  y := f.FFloat(y)

  return f.FSub(y, f.FMul(f_m, x))
  